import inspect
import json
from collections.abc import AsyncIterator, Callable, Coroutine, Iterator
from typing import Any, Protocol

from bolinette.core.injection import Injection
from bolinette.core.mapping import JsonObjectEncoder
from bolinette.web.abstract import Response
from bolinette.web.resources import HttpHeaders, ResponseData


class ResponseWriter:
    def __init__(self, inject: Injection, response: Response) -> None:
        self.inject = inject
        self.response = response

    async def close(self) -> None:
        await self.response.close()

    async def _open_response(self, data: ResponseData) -> None:
        self.response.set_status(data.status)
        for header, value in data.headers.items():
            self.response.set_header(header, value)
        await self.response.open()

    async def write_result(self, result: object, data: ResponseData) -> None:
        await self._unpack_result(result, data)

    async def _unpack_result(self, result: object, data: ResponseData) -> None:
        if inspect.iscoroutine(result) or inspect.isawaitable(result):
            return await self._unpack_result(await result, data)
        if inspect.isasyncgen(result):
            return await self._write_async_iterator(result, data)
        if inspect.isgenerator(result):
            return await self._write_iterator(result, data)
        if inspect.iscoroutinefunction(result):
            return await self._unpack_result(await self.inject.call(result), data)
        if inspect.isasyncgenfunction(result) or callable(result):
            return await self._unpack_result(self.inject.call(result), data)
        await self._write_single(result, data)

    @staticmethod
    def _get_value_writer(value: Any, data: ResponseData, as_list: bool) -> "ValueWriter[Any]":
        value_writer: ValueWriter[Any] | None = None
        match value:
            case None | bytes():
                value_writer = RawValueTransformer()
            case str():
                value_writer = StringValueTransformer()
            case _ if as_list:
                value_writer = JsonListValueTransformer()
            case _:
                value_writer = JsonValueTransformer()
        if not data.has_header(HttpHeaders.ContentType):
            data.set_header(HttpHeaders.ContentType, value_writer.default_content_type())
        return value_writer

    async def _write_async_iterator(self, result: AsyncIterator[object], data: ResponseData) -> None:
        value_writer: ValueWriter[Any] | None = None
        try:
            chunk = await anext(result)
            value_writer = self._get_value_writer(chunk, data, True)
            await self._open_response(data)
            await self._write_chunk(chunk, value_writer)
            while True:
                chunk = await anext(result)
                await self._write_chunk(chunk, value_writer)
        except StopAsyncIteration:
            pass
        finally:
            if value_writer is not None:
                await self._close_writer(value_writer)

    async def _write_iterator(self, result: Iterator[object], data: ResponseData) -> None:
        value_writer: ValueWriter[Any] | None = None
        try:
            chunk = next(result)
            value_writer = self._get_value_writer(chunk, data, True)
            await self._open_response(data)
            await self._write_chunk(chunk, value_writer)
            while True:
                chunk = next(result)
                await self._write_chunk(chunk, value_writer)
        except StopIteration:
            pass
        finally:
            if value_writer is not None:
                await self._close_writer(value_writer)

    async def _write_single(self, result: Any, data: ResponseData) -> None:
        value_writer = self._get_value_writer(result, data, False)
        await self._open_response(data)
        await self._write_chunk(result, value_writer)
        await self._close_writer(value_writer)

    def _write_chunk(self, value: object, value_writer: "ValueWriter[object]") -> Coroutine[Any, Any, None]:
        return value_writer.write(self.response.write, value)

    def _close_writer(self, value_writer: "ValueWriter[object]") -> Coroutine[Any, Any, None]:
        return value_writer.close(self.response.write)


class ValueWriter[T](Protocol):
    def default_content_type(self) -> str: ...
    async def write(self, write: Callable[[bytes], Coroutine[Any, Any, None]], value: T) -> None: ...
    async def close(self, write: Callable[[bytes], Coroutine[Any, Any, None]]) -> None: ...


class RawValueTransformer:
    def default_content_type(self) -> str:
        return "application/octet-stream"

    def write(self, write: Callable[[bytes], Coroutine[Any, Any, None]], value: bytes) -> Coroutine[Any, Any, None]:
        return write(value)

    async def close(self, write: Callable[[bytes], Coroutine[Any, Any, None]]) -> None:
        pass


class StringValueTransformer:
    def default_content_type(self) -> str:
        return "text/plain"

    def write(self, write: Callable[[bytes], Coroutine[Any, Any, None]], value: str) -> Coroutine[Any, Any, None]:
        return write(value.encode())

    async def close(self, write: Callable[[bytes], Coroutine[Any, Any, None]]) -> None:
        pass


class JsonValueTransformer:
    def __init__(self) -> None:
        self.item = 0

    def default_content_type(self) -> str:
        return "application/json"

    def write(self, write: Callable[[bytes], Coroutine[Any, Any, None]], value: Any) -> Coroutine[Any, Any, None]:
        return write(json.dumps(value, cls=JsonObjectEncoder, separators=(", ", ": ")).encode())

    async def close(self, write: Callable[[bytes], Coroutine[Any, Any, None]]) -> None:
        pass


class JsonListValueTransformer:
    def __init__(self) -> None:
        self.item = 0

    def default_content_type(self) -> str:
        return "application/json"

    async def write(self, write: Callable[[bytes], Coroutine[Any, Any, None]], value: Any) -> None:
        if self.item == 0:
            await write(b"[")
        else:
            await write(b", ")
        await write(json.dumps(value, cls=JsonObjectEncoder, separators=(", ", ": ")).encode())
        self.item += 1

    def close(self, write: Callable[[bytes], Coroutine[Any, Any, None]]) -> Coroutine[Any, Any, None]:
        return write(b"]")



from bolinette.web.asgi.types import (
    HttpReceivedEvent as HttpReceivedEvent,
    HttpRequestEvent as HttpRequestEvent,
    HttpResponseResult as HttpResponseResult,
    LifespanReceivedEvent as LifespanReceivedEvent,
    LifespanResult as LifespanResult,
    LifespanShutdownResult as LifespanShutdownResult,
    LifespanStartupResult as LifespanStartupResult,
    Scope as Scope,
    HttpScope as HttpScope,
    WebSocketScope as WebSocketScope,
    AsgiCallable as AsgiCallable,
    WebSocketReceivedEvent as WebSocketReceivedEvent,
    WebSocketSendResult as WebSocketSendResult,
    WebSocketConnectResult as WebSocketConnectResult,
    WebSocketResult as WebSocketResult,
)
from bolinette.web.asgi.requests import AsgiRequest as AsgiRequest, AsgiSocketRequest as AsgiSocketRequest
from bolinette.web.asgi.responses import AsgiSocketResponse as AsgiSocketResponse, AsgiResponse as AsgiResponse
from bolinette.web.asgi.app import AsgiApplication as AsgiApplication



import inspect
import json
from typing import Any, TypeGuard

from bolinette.core import Cache, Logger, meta
from bolinette.core.environment import CoreSection
from bolinette.core.injection import Injection, init_method
from bolinette.core.mapping.json import JsonObjectEncoder
from bolinette.core.types import Type, TypeChecker
from bolinette.core.utils import AttributeUtils
from bolinette.web.abstract import WebSocketRequest, WebSocketResponse
from bolinette.web.exceptions import BadRequestError, NotFoundError, WebErrorHandler
from bolinette.web.ws import ChannelMessage, WebSocketSubscription, WebSocketTopic
from bolinette.web.ws.channel import WebSocketChannelMeta
from bolinette.web.ws.requests import (
    ChannelRequest,
    ChannelSendRequest,
    ChannelSubscribeRequest,
    ChannelUnsubscribeRequest,
    SocketContent,
)
from bolinette.web.ws.topic import WebSocketTopicMeta


class WebSocketHandler:
    def __init__(
        self,
        logger: "Logger[WebSocketHandler]",
        checker: TypeChecker,
        inject: Injection,
        core_section: CoreSection,
    ) -> None:
        self.logger = logger
        self.checker = checker
        self.topics: dict[str, _WSTypeBag]
        self.inject = inject
        self.core_section = core_section
        self.subscriptions: dict[WebSocketResponse, dict[str, set[str]]] = {}

    @init_method
    def _init_topics(self, cache: Cache) -> None:
        self.topics = {}
        for cls in cache.get(WebSocketTopic, hint=type[WebSocketTopic[...]], raises=False):
            topic_meta = meta.get(cls, WebSocketTopicMeta)
            self.add_topic(topic_meta.name, cls)

    @init_method
    def _add_context_to_inject(self) -> None:
        self.inject.add(WebSocketContext, "singleton", [self])

    def add_topic(self, name: str, cls: type[WebSocketTopic[...]]) -> None:
        self.topics[name] = _WSTypeBag(Type(cls))

    async def handle(self, request: WebSocketRequest, response: WebSocketResponse) -> None:
        try:
            content = request.json()
            if not self.is_message(content):
                raise BadRequestError(
                    "Invalid message, action must be sub, unsub, send or close",
                    "ws.bad_request",
                )
            match content["action"]:
                case "sub":
                    if not self.checker.instanceof(content, ChannelSubscribeRequest):
                        raise BadRequestError(
                            "Invalid sub action, must contain topic and channel",
                            "ws.bad_request",
                        )
                    await self._subscribe(content, response)
                case "unsub":
                    if not self.checker.instanceof(content, ChannelUnsubscribeRequest):
                        raise BadRequestError(
                            "Invalid unsub action, must contain topic and channel",
                            "ws.bad_request",
                        )
                    await self._unsubscribe(content, response)
                case "send":
                    if not self.checker.instanceof(content, ChannelSendRequest[SocketContent]):
                        raise BadRequestError(
                            "Invalid send action, must contain topic, channel and data",
                            "ws.bad_request",
                        )
                    await self._send(content, response)
                case "close":
                    await self._close(response)
        except Exception as err:
            self.logger.error(str(type(err)), str(err))
            _, content = WebErrorHandler.create_error_payload(err, self.core_section.debug)
            await response.send(json=content)

    async def remove_connection(self, response: WebSocketResponse) -> None:
        await self._close(response)

    async def _subscribe(self, request: ChannelSubscribeRequest, response: WebSocketResponse) -> None:
        topic_name = request["topic"]
        channel_name = request["channel"]
        if topic_name not in self.topics:
            raise NotFoundError("Unknown topic", "ws.topic.not_found", {"topic": topic_name})
        topic = self.topics[topic_name]
        async with self.inject.get_async_scoped_session() as subinject:
            topic_instance = subinject.instantiate(topic.t.cls)
            result = await topic_instance.subscribe(WebSocketSubscription(channel_name))
            if not result:
                raise Exception()  # TODO
            topic.add_subscription(channel_name, response)
            self._add_subscription(response, topic_name, channel_name)

    async def _unsubscribe(self, request: ChannelUnsubscribeRequest, response: WebSocketResponse) -> None:
        topic_name = request["topic"]
        channel_name = request["channel"]
        if topic_name not in self.topics:
            raise Exception()  # TODO
        topic = self.topics[topic_name]
        topic.remove_subscription(channel_name, response)
        self._remove_subscription(response, topic_name, channel_name)

    async def _send(self, request: ChannelSendRequest[SocketContent], response: WebSocketResponse) -> None:
        topic_name = request["topic"]
        if topic_name not in self.topics:
            raise NotFoundError("Unknown topic", "ws.topic.not_found", {"topic": topic_name})
        topic = self.topics[topic_name]
        if (channel := topic.match(request["channel"])) is not None:
            async with self.inject.get_async_scoped_session() as subinject:
                topic_instance = subinject.instantiate(topic.t.cls)
                message = ChannelMessage(request["channel"], request["data"], response)
                if inspect.isawaitable(res := subinject.call(channel.func, args=[topic_instance, message])):
                    await res

    async def _close(self, response: WebSocketResponse) -> None:
        if response in self.subscriptions:
            for topic_name, channels in self.subscriptions[response].items():
                for channel_name in channels:
                    self.topics[topic_name].remove_subscription(channel_name, response)
            del self.subscriptions[response]

    def _add_subscription(self, ws: WebSocketResponse, topic: str, channel: str) -> None:
        if ws not in self.subscriptions:
            self.subscriptions[ws] = {}
        resp_subs = self.subscriptions[ws]
        if topic not in resp_subs:
            resp_subs[topic] = set()
        resp_subs[topic].add(channel)

    def _remove_subscription(self, ws: WebSocketResponse, topic: str, channel: str) -> None:
        if ws not in self.subscriptions:
            raise KeyError()
        resp_subs = self.subscriptions[ws]
        if topic not in resp_subs:
            raise KeyError()
        resp_subs[topic].remove(channel)

    @staticmethod
    def is_message(content: Any) -> TypeGuard[ChannelRequest]:
        return (
            isinstance(content, dict)
            and ("action" in content)
            and (content["action"] in ("sub", "unsub", "send", "close"))
        )


class _WSTypeBag:
    def __init__(self, cls: "Type[WebSocketTopic[...]]") -> None:
        self.t = cls
        self.subs: dict[str, set[WebSocketResponse]] = {}

    def match(self, channel: str) -> WebSocketChannelMeta[Any, Any, ..., Any] | None:
        for attr in AttributeUtils.get_cls_attrs(self.t.cls).values():
            if meta.has(attr, WebSocketChannelMeta):
                chan_meta: WebSocketChannelMeta[Any, Any, ..., Any] = meta.get(attr, WebSocketChannelMeta)
                if chan_meta.pattern.match(channel):
                    return chan_meta
        return None

    def add_subscription(self, channel: str, ws: WebSocketResponse) -> None:
        if channel not in self.subs:
            self.subs[channel] = {ws}
        else:
            self.subs[channel].add(ws)

    def remove_subscription(self, channel: str, ws: WebSocketResponse) -> None:
        if channel not in self.subs:
            raise KeyError()  # TODO
        subs = self.subs[channel]
        if ws not in subs:
            raise ValueError()
        subs.remove(ws)

    def is_registered(self, channel: str, ws: WebSocketResponse) -> bool:
        return channel in self.subs and ws in self.subs[channel]


class WebSocketContext:
    def __init__(self, handler: WebSocketHandler) -> None:
        self._handler = handler

    async def send(self, topic: str, channel: str, content: SocketContent) -> None:
        if topic not in self._handler.topics:
            return
        topic_t = self._handler.topics[topic]
        if channel not in topic_t.subs:
            return
        for ws in topic_t.subs[channel]:
            await ws.send(json=content, encoder=lambda o: json.dumps(o, cls=JsonObjectEncoder))



import re
from collections.abc import Callable
from typing import Literal

from bolinette.core import meta

HttpMethod = Literal["GET", "POST", "PUT", "PATCH", "DELETE"]
path_re = re.compile(r"\{([^}]+)\}")


class RouteProps:
    def __init__(self, method: HttpMethod, path: str) -> None:
        self.method = method
        self.path = path


class RouteBucket:
    def __init__(self, name: str) -> None:
        self.name = name
        self.routes: list[RouteProps] = []

    def add(self, route: RouteProps) -> None:
        self.routes.append(route)

    def __getitem__(self, index: int) -> RouteProps:
        return self.routes[index]


def route[**FuncP, FuncT](
    method: HttpMethod,
    path: str,
    /,
) -> Callable[[Callable[FuncP, FuncT]], Callable[FuncP, FuncT]]:
    def decorator(func: Callable[FuncP, FuncT]) -> Callable[FuncP, FuncT]:
        if not meta.has(func, RouteBucket):
            bucket = RouteBucket(func.__name__)
            meta.set(func, bucket)
        else:
            bucket = meta.get(func, RouteBucket)
        bucket.add(RouteProps(method, path))
        return func

    return decorator


def get[**FuncP, FuncT](path: str, /) -> Callable[[Callable[FuncP, FuncT]], Callable[FuncP, FuncT]]:
    return route("GET", path)


def post[**FuncP, FuncT](path: str, /) -> Callable[[Callable[FuncP, FuncT]], Callable[FuncP, FuncT]]:
    return route("POST", path)


def put[**FuncP, FuncT](path: str, /) -> Callable[[Callable[FuncP, FuncT]], Callable[FuncP, FuncT]]:
    return route("PUT", path)


def patch[**FuncP, FuncT](path: str, /) -> Callable[[Callable[FuncP, FuncT]], Callable[FuncP, FuncT]]:
    return route("PATCH", path)


def delete[**FuncP, FuncT](path: str, /) -> Callable[[Callable[FuncP, FuncT]], Callable[FuncP, FuncT]]:
    return route("DELETE", path)



import os
import re
from typing import Any, override

from bolinette.core.types import Function, Type
from bolinette.web import Controller


class Route[**FuncP, FuncT]:
    def __init__(
        self,
        method: str,
        path: str,
        controller: Type[Controller],
        func: Function[FuncP, FuncT],
    ) -> None:
        self.method = method
        self.controller = controller
        self.func = func
        self.path = path

    def __call__(self, *args: FuncP.args, **kwargs: FuncP.kwargs) -> FuncT:
        return self.func(*args, **kwargs)

    @override
    def __repr__(self) -> str:
        return f"<Route {self.method} {self.path} -> {self.func}>"


class Resource:
    def __init__(self, path: str, routes: list[Route[..., Any]] | None = None, /) -> None:
        self.path = path
        self.routes = {r.method: r for r in routes or []}

    def add(self, route: Route[..., Any], /) -> None:
        if route.path != self.path:
            raise ValueError()  # TODO
        self.routes[route.method] = route

    def __getitem__(self, method: str, /) -> Route[..., Any]:
        return self.routes[method]

    def __contains__(self, method: str, /) -> bool:
        return method in self.routes

    @override
    def __repr__(self) -> str:
        return f"<Resource {self.path}: {len(self.routes)} routes>"

    def __or__(self, res: "Resource") -> "Resource":
        if self.path != res.path:
            raise ValueError("Resources must have the same path to be merged")
        new_res = Resource(self.path)
        new_res.routes = self.routes | res.routes
        return new_res


class ResourceNode:
    _PATH_PARAM_REGEX = re.compile(r"\{([^}:]+)(?:\:([^}]+))?\}")

    def __init__(
        self,
        resource: Resource | None,
        subnodes: "list[ResourceNode]",
    ) -> None:
        self.resource = resource
        self.subnodes = subnodes

    def match(self, path: str, path_params: dict[str, str]) -> int:
        raise NotImplementedError()

    def set_resource(self, resource: Resource | None, /) -> None:
        self.resource = resource

    def add_subnode(self, node: "ResourceNode", /) -> None:
        self.subnodes.append(node)

    @staticmethod
    def from_resource(resource: Resource) -> "ResourceNode":
        origin = resource.path
        first_node: ResourceNode | None = None
        last_node: ResourceNode | None = None
        last_index = 0
        for match in ResourceNode._PATH_PARAM_REGEX.finditer(origin):
            pattern_node = PatternResourceNode(match.group(1), match.group(2), None, None, [])
            static_node = StaticResourceNode(origin[last_index : match.start()], None, [pattern_node])
            if last_node is not None:
                last_node.add_subnode(static_node)
            if first_node is None:
                first_node = static_node
            last_node = pattern_node
            last_index = match.end()
        if last_index < len(origin):
            static_node = StaticResourceNode(origin[last_index:], resource, [])
            if last_node is not None:
                last_node.add_subnode(static_node)
            if first_node is None:
                first_node = static_node
        elif last_node is not None:
            last_node.resource = resource
        if first_node is None:
            raise ValueError("Empty path")
        return first_node

    @staticmethod
    def merge(n1: "ResourceNode", n2: "ResourceNode") -> "ResourceNode":
        def merge_resources(r1: Resource | None, r2: Resource | None) -> Resource | None:
            if r1 is None and r2 is None:
                return None
            if r1 is not None and r2 is None:
                return r1
            if r1 is None and r2 is not None:
                return r2
            if r1 is not None and r2 is not None:
                return r1 | r2

        if isinstance(n1, PatternResourceNode) and isinstance(n2, PatternResourceNode):
            if n1.param_name != n2.param_name and n1.norm_pattern != n2.norm_pattern:
                raise ValueError("Cannot merge different pattern")
            return PatternResourceNode(
                n1.param_name,
                n1.pattern,
                n1.norm_pattern,
                merge_resources(n1.resource, n2.resource),
                ResourceNode.merge_subnodes([*n1.subnodes, *n2.subnodes]),
            )

        if not isinstance(n1, StaticResourceNode) or not isinstance(n2, StaticResourceNode):
            raise TypeError("Cannot merge pattern path nodes")
        if n1.path == n2.path:
            return StaticResourceNode(
                n1.path,
                merge_resources(n1.resource, n2.resource),
                ResourceNode.merge_subnodes([*n1.subnodes, *n2.subnodes]),
            )
        common_path = os.path.commonprefix([n1.path, n2.path])
        if common_path == n1.path:
            return StaticResourceNode(
                common_path,
                n1.resource,
                ResourceNode.merge_subnodes(
                    [
                        *n1.subnodes,
                        StaticResourceNode(n2.path[len(common_path) :], n2.resource, n2.subnodes),
                    ]
                ),
            )
        elif common_path == n2.path:
            return StaticResourceNode(
                common_path,
                n2.resource,
                ResourceNode.merge_subnodes(
                    [
                        StaticResourceNode(n1.path[len(common_path) :], n1.resource, n1.subnodes),
                        *n2.subnodes,
                    ]
                ),
            )
        else:
            return StaticResourceNode(
                common_path,
                None,
                [
                    StaticResourceNode(n1.path[len(common_path) :], n1.resource, n1.subnodes),
                    StaticResourceNode(n2.path[len(common_path) :], n2.resource, n2.subnodes),
                ],
            )

    @staticmethod
    def merge_subnodes(nodes: "list[ResourceNode]") -> "list[ResourceNode]":
        merged_nodes: list[ResourceNode] = []
        for node in nodes:
            closest_match: tuple[str, int, ResourceNode] | None = None
            if isinstance(node, StaticResourceNode):
                for node_index, merged_node in enumerate(merged_nodes):
                    if not isinstance(merged_node, StaticResourceNode):
                        continue
                    common_path = os.path.commonprefix([node.path, merged_node.path])
                    if not len(common_path):
                        continue
                    if closest_match is None or len(common_path) > len(closest_match[0]):
                        closest_match = (common_path, node_index, merged_node)
            elif isinstance(node, PatternResourceNode):
                for node_index, merged_node in enumerate(merged_nodes):
                    if not isinstance(merged_node, PatternResourceNode):
                        continue
                    if node.param_name == merged_node.param_name and node.norm_pattern == merged_node.norm_pattern:
                        closest_match = ("", node_index, merged_node)
            if closest_match is None:
                merged_nodes.append(node)
            else:
                merged_nodes[closest_match[1]] = ResourceNode.merge(node, closest_match[2])
        return merged_nodes


class StaticResourceNode(ResourceNode):
    def __init__(
        self,
        path: str,
        resource: Resource | None,
        subnodes: list[ResourceNode],
    ) -> None:
        super().__init__(resource, subnodes)
        self.path = path

    @override
    def match(self, path: str, path_params: dict[str, str]) -> int:
        if path.startswith(self.path):
            return len(self.path)
        return 0

    @override
    def __repr__(self) -> str:
        return (
            f"<ResourceNode {self.path} "
            f"{len(self.resource.routes) if self.resource else 0} routes, {len(self.subnodes)} subnodes>"
        )


class PatternResourceNode(ResourceNode):
    def __init__(
        self,
        param_name: str,
        pattern: str | None,
        norm_pattern: str | None,
        resource: Resource | None,
        subnodes: list[ResourceNode],
    ) -> None:
        super().__init__(resource, subnodes)
        self.param_name = param_name
        self.pattern = pattern
        if norm_pattern is not None:
            self.norm_pattern = norm_pattern
        else:
            self.norm_pattern = pattern or "[^/]+"
        self.regex = re.compile(f"({self.norm_pattern})")

    @override
    def match(self, path: str, path_params: dict[str, str]) -> int:
        if match := self.regex.match(path):
            group = match.group(1)
            path_params[self.param_name] = group
            return len(group)
        return 0

    @override
    def __repr__(self) -> str:
        return (
            f"<ResourceNode {self.param_name}:{self.norm_pattern} "
            f"{len(self.resource.routes) if self.resource else 0} routes, {len(self.subnodes)} subnodes>"
        )



class WebSocketSubResult:
    def __init__(self, accepted: bool) -> None:
        self._accepted = accepted

    def __bool__(self) -> bool:
        return self._accepted


class WebSocketSubscription:
    def __init__(self, channel: str) -> None:
        self.channel = channel

    def accept(self) -> WebSocketSubResult:
        return WebSocketSubResult(True)

    def reject(self) -> WebSocketSubResult:
        return WebSocketSubResult(False)



from bolinette.web.routing.decorators import (
    route as route,
    get as get,
    post as post,
    put as put,
    patch as patch,
    delete as delete,
    HttpMethod as HttpMethod,
    RouteProps as RouteProps,
    RouteBucket as RouteBucket,
)
from bolinette.web.routing.resource import Resource as Resource, Route as Route, ResourceNode as ResourceNode
from bolinette.web.routing.router import Router as Router



import traceback
from http import HTTPStatus
from typing import Any, NotRequired, TypedDict, override

from bolinette.core.exceptions import BolinetteError, ParameterError
from bolinette.core.expressions import ExpressionNode, ExpressionTree
from bolinette.core.types import Function, Type
from bolinette.web.controller import Controller


class WebError(BolinetteError, ParameterError):
    def __init__(
        self,
        message: str,
        error_code: str,
        status: HTTPStatus,
        error_args: dict[str, Any] | None = None,
        *,
        ctrl: Type[Controller] | None = None,
        route: Function[..., Any] | None = None,
    ) -> None:
        ParameterError.__init__(self, ctrl="Controller {}", route="Route {}")
        BolinetteError.__init__(self, self._format_params(message, ctrl=ctrl, route=route))
        self.error_code = error_code
        self.status = status
        self.error_args = error_args or {}

    @override
    def __str__(self) -> str:
        if not self.error_args:
            return self.error_code
        return f"{self.error_code}|{'|'.join(f'{k}:{v}' for k,v in self.error_args.items())}"


class GroupedWebError(WebError):
    def __init__(
        self,
        errors: list[WebError],
        status: HTTPStatus,
        *,
        ctrl: Type[Controller] | None = None,
        route: Function[..., Any] | None = None,
    ) -> None:
        super().__init__(
            "Error raised during route executions",
            "internal.web.exception",
            status,
            {},
            ctrl=ctrl,
            route=route,
        )
        self.errors = errors

    def add(self, *errors: WebError) -> None:
        self.errors.extend(errors)


class BadRequestError(WebError):
    def __init__(
        self,
        message: str,
        error_code: str,
        error_args: dict[str, Any] | None = None,
        *,
        ctrl: Type[Controller] | None = None,
        route: Function[..., Any] | None = None,
    ) -> None:
        super().__init__(message, error_code, HTTPStatus.BAD_REQUEST, error_args, ctrl=ctrl, route=route)


class MissingParameterError(BadRequestError):
    def __init__(
        self,
        expr: ExpressionNode,
        *,
        ctrl: Type[Controller] | None = None,
        route: Function[..., Any] | None = None,
    ) -> None:
        path = ExpressionTree.format(expr, max_depth=1)
        super().__init__(
            f"Parameter '{path}' is missing in payload",
            "payload.parameter.missing",
            {"path": path},
            ctrl=ctrl,
            route=route,
        )


class ParameterNotNullableError(BadRequestError):
    def __init__(
        self,
        expr: ExpressionNode,
        *,
        ctrl: Type[Controller] | None = None,
        route: Function[..., Any] | None = None,
    ) -> None:
        path = ExpressionTree.format(expr, max_depth=1)
        super().__init__(
            f"Parameter '{path}' must not be null",
            "payload.parameter.not_nullable",
            {"path": path},
            ctrl=ctrl,
            route=route,
        )


class WrongParameterTypeError(BadRequestError):
    def __init__(
        self,
        expr: ExpressionNode,
        target: Type[Any],
        *,
        ctrl: Type[Controller] | None = None,
        route: Function[..., Any] | None = None,
    ) -> None:
        path = ExpressionTree.format(expr, max_depth=1)
        super().__init__(
            f"Parameter '{path}' could be converted to '{target}'",
            "payload.parameter.wrong_type",
            {"path": path},
            ctrl=ctrl,
            route=route,
        )


class NotFoundError(WebError):
    def __init__(
        self,
        message: str,
        error_code: str,
        error_args: dict[str, Any] | None = None,
        *,
        ctrl: Type[Controller] | None = None,
        route: Function[..., Any] | None = None,
    ) -> None:
        super().__init__(message, error_code, HTTPStatus.NOT_FOUND, error_args, ctrl=ctrl, route=route)


class InternalServerError(WebError):
    def __init__(
        self,
        message: str,
        error_code: str | None = None,
        error_args: dict[str, Any] | None = None,
        *,
        ctrl: Type[Controller] | None = None,
        route: Function[..., Any] | None = None,
    ) -> None:
        super().__init__(
            message,
            error_code or "internal.web.exception",
            HTTPStatus.INTERNAL_SERVER_ERROR,
            error_args,
            ctrl=ctrl,
            route=route,
        )


class DispatchError(BolinetteError, ParameterError):
    def __init__(
        self,
        message: str,
        *,
        route: str | None = None,
    ) -> None:
        ParameterError.__init__(self, route="Route {}")
        BolinetteError.__init__(self, self._format_params(message, route=route))


class NotFoundDispatchError(DispatchError):
    def __init__(self, route: str) -> None:
        super().__init__("Route not found", route=route)


class MethodNotAllowedDispatchError(DispatchError):
    def __init__(self, route: str) -> None:
        super().__init__("Route not found", route=route)


class WebErrorHandler:
    @classmethod
    def create_error_payload(cls, err: Exception, debug: bool) -> "tuple[HTTPStatus, ErrorResponseContent]":
        if isinstance(err, WebError):
            status = err.status
            if isinstance(err, GroupedWebError):
                errors = err.errors
            else:
                errors = [err]
        else:
            status = HTTPStatus.INTERNAL_SERVER_ERROR
            errors = [err]
        content: ErrorResponseContent = {
            "status": status.value,
            "reason": status.phrase,
            "errors": [cls._format_error(e) for e in errors],
        }
        if debug:
            content["debug"] = {
                "type": str(type(err)),
                "message": str(err),
                "stacktrace": traceback.format_exc().split("\n"),
            }
        return status, content

    @staticmethod
    def _format_error(err: Exception) -> "ErrorDescription":
        if isinstance(err, WebError):
            message = err.message
            code = err.error_code
            params = err.error_args
        else:
            message = "Un unexpected error has occured while processing the request"
            code = "internal.error"
            params = {}
        f_err: ErrorDescription = {"message": message, "code": code, "params": params}
        return f_err


class DebugErrorDetails(TypedDict):
    type: str
    message: str
    stacktrace: list[str]


class ErrorDescription(TypedDict):
    message: str
    code: str
    params: dict[str, Any]


class ErrorResponseContent(TypedDict):
    status: int
    reason: str
    errors: list[ErrorDescription]
    debug: NotRequired[DebugErrorDetails]



import json
from collections.abc import Awaitable, Callable
from typing import Any, Literal

from bolinette.web.asgi import HttpReceivedEvent, HttpRequestEvent


class AsgiAsyncBody:
    def __init__(
        self,
        received: HttpRequestEvent,
        receive: Callable[[], Awaitable[HttpReceivedEvent]],
    ) -> None:
        self.chunks = [received.get("body", b"")]
        self.more = received.get("more_body", False)
        self.receive = receive

    async def read(self) -> bytes:
        while self.more:
            received = await self.receive()
            self.chunks.append(received.get("body", b""))
            self.more = received.get("more_body", False)
        return b"".join(self.chunks)


class AsgiRequest:
    def __init__(
        self,
        method: str,
        path: str,
        headers: dict[str, str],
        query_params: dict[str, str],
        received: HttpRequestEvent,
        receive: Callable[[], Awaitable[HttpReceivedEvent]],
    ) -> None:
        self.method = method
        self.path = path
        self.headers = headers
        self.query_params = query_params
        self.path_params: dict[str, str] = {}
        self._body = AsgiAsyncBody(received, receive)

    async def raw(self) -> bytes:
        return await self._body.read()

    async def text(self, *, encoding: str = "utf-8") -> str:
        return (await self._body.read()).decode(encoding)

    async def json(self, *, cls: type[json.JSONDecoder] | None = None) -> Any:
        return json.loads(await self._body.read(), cls=cls)


class AsgiSocketRequest:
    def __init__(
        self,
        bytes: bytes | None,
        text: str | None,
    ) -> None:
        self._bytes = bytes
        self._text = text

    def get_type(self) -> Literal["raw", "text"]:
        if self._bytes is not None:
            return "raw"
        return "text"

    def raw(self) -> bytes:
        if self._bytes is None:
            raise TypeError("This request has a unicode content")
        return self._bytes

    def text(self) -> str:
        if self._text is None:
            raise TypeError("This request has a raw content")
        return self._text

    def json(self, *, cls: type[json.JSONDecoder] | None = None) -> Any:
        if self._bytes is not None:
            return json.loads(self._bytes, cls=cls)
        if self._text is not None:
            return json.loads(self._text, cls=cls)



from collections.abc import Awaitable, Callable
from typing import Any, Literal, NotRequired, TypedDict


class ASGIVersion(TypedDict):
    version: str
    spec_version: NotRequired[str]


class LifespanScope(TypedDict):
    type: Literal["lifespan"]
    asgi: ASGIVersion


class WebMessageScope(TypedDict):
    asgi: ASGIVersion
    http_version: Literal["1.0", "1.1", "2", "3"]
    server: NotRequired[tuple[str, int] | tuple[str, None]]
    client: NotRequired[tuple[str, int]]
    scheme: NotRequired[str]
    method: str
    root_path: NotRequired[str]
    path: str
    raw_path: NotRequired[bytes]
    query_string: bytes
    headers: list[tuple[bytes, bytes]]
    state: NotRequired[dict[str, Any]]


class HttpScope(WebMessageScope):
    type: Literal["http"]


class WebSocketScope(WebMessageScope):
    type: Literal["websocket"]
    subprotocols: NotRequired[list[str]]


class LifespanStartupEvent(TypedDict):
    type: Literal["lifespan.startup"]


class LifespanShutdownEvent(TypedDict):
    type: Literal["lifespan.shutdown"]


class LifespanStartupComplete(TypedDict):
    type: Literal["lifespan.startup.complete"]


class LifespanStartupFailed(TypedDict):
    type: Literal["lifespan.startup.failed"]


class LifespanShutdownComplete(TypedDict):
    type: Literal["lifespan.shutdown.complete"]


class LifespanShutdownFailed(TypedDict):
    type: Literal["lifespan.shutdown.failed"]


class HttpRequestEvent(TypedDict):
    type: Literal["http.request"]
    body: NotRequired[bytes]
    more_body: NotRequired[bool]


class HttpDisconectEvent(TypedDict):
    type: Literal["http.disconnect"]


class HttpResponseStart(TypedDict):
    type: Literal["http.response.start"]
    status: int
    headers: NotRequired[list[tuple[bytes, bytes]]]
    trailers: NotRequired[bool]


class HttpResponseBody(TypedDict):
    type: Literal["http.response.body"]
    body: NotRequired[bytes]
    more_body: NotRequired[bool]


class WebSocketConnectEvent(TypedDict):
    type: Literal["websocket.connect"]


class WebSocketReceiveBytesEvent(TypedDict):
    type: Literal["websocket.receive"]
    bytes: bytes
    text: NotRequired[None]


class WebSocketReceiveTextEvent(TypedDict):
    type: Literal["websocket.receive"]
    bytes: NotRequired[None]
    text: str


class WebSocketDisconnectEvent(TypedDict):
    type: Literal["websocket.disconnect"]
    code: int


class WebSocketAcceptResult(TypedDict):
    type: Literal["websocket.accept"]
    subprotocol: NotRequired[str]
    headers: NotRequired[list[tuple[bytes, bytes]]]


class WebSocketSendBytesResult(TypedDict):
    type: Literal["websocket.send"]
    bytes: bytes
    text: NotRequired[None]


class WebSocketSendTextResult(TypedDict):
    type: Literal["websocket.send"]
    bytes: NotRequired[None]
    text: str


class WebSocketCloseResult(TypedDict):
    type: Literal["websocket.close"]
    code: NotRequired[int]
    reason: NotRequired[str | None]


type Scope = LifespanScope | HttpScope | WebSocketScope

type LifespanReceivedEvent = LifespanStartupEvent | LifespanShutdownEvent
type LifespanStartupResult = LifespanStartupComplete | LifespanStartupFailed
type LifespanShutdownResult = LifespanShutdownComplete | LifespanShutdownFailed
type LifespanResult = LifespanStartupResult | LifespanShutdownResult

type HttpReceivedEvent = HttpRequestEvent | HttpDisconectEvent
type HttpResponseResult = HttpResponseStart | HttpResponseBody

type WebSocketReceivedEvent = (
    WebSocketConnectEvent | WebSocketReceiveBytesEvent | WebSocketReceiveTextEvent | WebSocketDisconnectEvent
)
type WebSocketConnectResult = WebSocketAcceptResult | WebSocketCloseResult
type WebSocketSendResult = WebSocketSendBytesResult | WebSocketSendTextResult
type WebSocketResult = WebSocketConnectResult | WebSocketSendResult

type AsgiCallable = Callable[[Scope, Callable[[], Awaitable[Any]], Callable[[Any], Awaitable[None]]], Awaitable[Any]]



from bolinette.web.resources.headers import HttpHeaders as HttpHeaders
from bolinette.web.resources.data import ResponseData as ResponseData
from bolinette.web.resources.resolvers import (
    RouteParamArgResolver as RouteParamArgResolver,
    RoutePayloadArgResolver as RoutePayloadArgResolver,
)
from bolinette.web.resources.writer import ResponseWriter as ResponseWriter
from bolinette.web.resources.resources import WebResources as WebResources



import json
from collections.abc import Callable
from enum import Enum, auto, unique
from typing import Any, Protocol, overload


@unique
class ResponseState(Enum):
    Idle = auto()  # Has not started yet
    Started = auto()  # Has sent headers, but no body yet
    Sending = auto()  # Has started streaming body
    Closed = auto()  # Has been closed


class Response(Protocol):
    @property
    def status(self) -> int: ...
    @property
    def headers(self) -> dict[str, str]: ...
    @property
    def state(self) -> ResponseState: ...

    async def open(self) -> None: ...

    async def close(self) -> None: ...

    async def write(self, raw: bytes) -> None: ...

    def set_status(self, status: int, /) -> None: ...

    def set_header(self, key: str, value: str, /) -> None: ...

    def has_header(self, key: str, /) -> bool: ...

    def unset_header(self, key: str, /) -> None: ...


class WebSocketResponse(Protocol):
    @overload
    async def send(self, *, raw: bytes) -> None: ...
    @overload
    async def send(self, *, text: str) -> None: ...
    @overload
    async def send(self, *, json: Any, encoder: Callable[[object], str] = json.dumps) -> None: ...



import json
from collections.abc import Callable
from http import HTTPStatus
from typing import Any

from bolinette.core import Cache, Logger, meta
from bolinette.core.environment import CoreSection
from bolinette.core.injection import Injection, ScopedInjection, init_method
from bolinette.core.mapping import Mapper
from bolinette.core.types import Function, Type, TypeChecker, TypeVarLookup
from bolinette.core.utils import AttributeUtils
from bolinette.web.abstract import Request, Response, ResponseState
from bolinette.web.config import WebConfig
from bolinette.web.controller import Controller, ControllerMeta
from bolinette.web.exceptions import (
    MethodNotAllowedDispatchError,
    NotFoundDispatchError,
    WebErrorHandler,
)
from bolinette.web.middleware import Middleware, MiddlewareBag
from bolinette.web.resources import (
    HttpHeaders,
    ResponseData,
    ResponseWriter,
    RouteParamArgResolver,
    RoutePayloadArgResolver,
)
from bolinette.web.routing import Route, RouteBucket, Router
from bolinette.web.ws import WebSocketHandler


class WebResources:
    def __init__(
        self,
        inject: Injection,
        logger: "Logger[WebResources]",
        core_section: CoreSection,
        checker: TypeChecker,
    ) -> None:
        self.inject = inject
        self.logger = logger
        self.core_section = core_section
        self.checker = checker
        self.router = Router()
        self.ws_handler: WebSocketHandler | None

    @init_method
    def _init_ctrls(self, cache: Cache) -> None:
        for ctrl_cls in cache.get(ControllerMeta, hint=type[Controller], raises=False):
            ctrl_meta = meta.get(ctrl_cls, ControllerMeta)
            attr: Any
            for attr in AttributeUtils.get_cls_attrs(ctrl_cls).values():
                if not meta.has(attr, RouteBucket):
                    continue
                bucket = meta.get(attr, RouteBucket)
                for props in bucket.routes:
                    self.add_route(ctrl_cls, ctrl_meta.path, attr, props.method, props.path)

    @init_method
    def _init_sockets(self, config: WebConfig, inject: Injection) -> None:
        if config.use_sockets:
            self.ws_handler = inject.add(WebSocketHandler, "singleton", instantiate=True)

    def add_route(
        self,
        ctrl_cls: type[Controller],
        controller_path: str,
        route_func: Callable[..., Any],
        method: str,
        route_path: str,
    ) -> None:
        if not route_path.startswith("/"):
            route_path = "/".join(p for p in [controller_path.removesuffix("/"), route_path] if p)
        if not route_path.startswith("/"):
            route_path = f"/{route_path}"
        self.router.add_route(Route(method, route_path, Type(ctrl_cls), Function(route_func)))

    async def dispatch(self, request: Request, response: Response) -> None:
        result: object | None = None
        data: ResponseData | None = None
        try:
            route = self.router.dispatch(request)
            await self._handle_request(route, request, response)
        except NotFoundDispatchError:
            result = "404 Not Found"
            data = ResponseData(status=HTTPStatus.NOT_FOUND, headers={HttpHeaders.ContentType: "text/plain"})
        except MethodNotAllowedDispatchError:
            result = "405 Method Not Allowed"
            data = ResponseData(status=HTTPStatus.METHOD_NOT_ALLOWED, headers={HttpHeaders.ContentType: "text/plain"})
        if result is not None and data is not None:
            if response.state != ResponseState.Idle:
                self.logger.error("Response has already started, unable to send error")
            else:
                writer = ResponseWriter(self.inject, response)
                await writer.write_result(result, data)
                await writer.close()

    @staticmethod
    def _prepare_session(inject: Injection, request: Request, data: ResponseData) -> None:
        inject.add(Request, "scoped", instance=request)
        inject.add(ResponseData, "scoped", instance=data)

    async def _handle_request(
        self,
        route: Route[..., Any],
        request: Request,
        response: Response,
    ) -> None:
        self.logger.info(f"Received request on {request.path}")
        writer = ResponseWriter(self.inject, response)
        try:
            async with self.inject.get_async_scoped_session() as scoped_inject:
                data = ResponseData()
                self._prepare_session(scoped_inject, request, data)
                mdlws = self._collect_middlewares(route, scoped_inject)
                result = await self._middleware_chain(route, request, scoped_inject, mdlws)
                await writer.write_result(result, data)
        except Exception as err:
            self.logger.error(str(type(err)), str(err))
            status, content = WebErrorHandler.create_error_payload(err, self.core_section.debug)
            if response.state != ResponseState.Idle:
                self.logger.error("Response has already started, unable to send error")
            else:
                await writer.write_result(
                    content,
                    ResponseData(
                        status=status,
                        headers={HttpHeaders.ContentType: "application/json"},
                    ),
                )
        finally:
            await writer.close()

    async def _middleware_chain(
        self,
        route: Route[..., Any],
        request: Request,
        scoped: ScopedInjection,
        mdlws: list[Middleware[Any]],
        index: int = 0,
    ) -> Any:
        async def _next_handle() -> Any:
            return await self._middleware_chain(route, request, scoped, mdlws, index + 1)

        if index >= len(mdlws):
            return await self._call_controller(route, request, scoped)
        else:
            mdlw = mdlws[index]
            self.logger.debug(f"Calling middleware {mdlw}")
            return await scoped.call(mdlw.handle, args=[_next_handle])

    async def _call_controller(
        self,
        route: Route[..., Any],
        request: Request,
        scoped: Injection,
    ) -> Any:
        try:
            body = await request.json()
        except json.JSONDecodeError:
            body = None
        ctrl = scoped.instantiate(route.controller.cls)
        self.logger.debug(f"Calling controller route {route.func}(...)")
        return scoped.call(
            route.func.func,
            args=[ctrl],
            additional_resolvers=[
                RouteParamArgResolver(route, request),
                RoutePayloadArgResolver(route, scoped.require(Mapper), body),
            ],
            vars_lookup=TypeVarLookup(route.controller),
        )

    @staticmethod
    def _collect_middlewares(route: Route[..., Any], scoped: Injection) -> list[Middleware[Any]]:
        bags: list[MiddlewareBag] = []
        if meta.has(route.controller.cls, MiddlewareBag):
            bags.append(meta.get(route.controller.cls, MiddlewareBag))
        if meta.has(route.func, MiddlewareBag):
            bags.append(meta.get(route.func, MiddlewareBag))
        mdlws: dict[Type[Middleware[...]], Middleware[...]] = {}
        for bag in bags:
            for t, mdlw_meta in reversed(bag.added.items()):
                mdlw = scoped.instantiate(t.cls)
                mdlw.options(*mdlw_meta.args, **mdlw_meta.kwargs)
                mdlws[t] = mdlw
            for t in bag.removed:
                if t in mdlws:
                    del mdlws[t]
        return list(mdlws.values())



import json
from typing import Any, Literal, Protocol


class Request(Protocol):
    method: str
    path: str
    headers: dict[str, str]
    query_params: dict[str, str]
    path_params: dict[str, str]

    async def raw(self) -> bytes: ...

    async def text(self, *, encoding: str = "utf-8") -> str: ...

    async def json(self, *, cls: type[json.JSONDecoder] | None = None) -> Any: ...


class WebSocketRequest(Protocol):
    def get_type(self) -> Literal["raw", "text"]: ...

    def raw(self) -> bytes: ...

    def text(self) -> str: ...

    def json(self, *, cls: type[json.JSONDecoder] | None = None) -> Any: ...



from http import HTTPStatus
from typing import Any, overload

from bolinette.web.resources import HttpHeaders


class ResponseData:
    def __init__(self, *, status: HTTPStatus = HTTPStatus.OK, headers: dict[str, str] | None = None) -> None:
        self._status = status
        self._headers = headers or {}

    @property
    def status(self) -> HTTPStatus:
        return self._status

    def set_status(self, value: HTTPStatus | int, /) -> None:
        if isinstance(value, HTTPStatus):
            self._status = value
        else:
            self._status = HTTPStatus(value)

    @property
    def headers(self) -> dict[str, str]:
        return {**self._headers}

    @overload
    def get_header(self, key: str, /) -> str: ...
    @overload
    def get_header[T](self, key: str, default: T, /) -> str | T: ...

    def get_header(self, key: str, /, *args: Any) -> Any:
        if len(args):
            return self._headers.get(key, args[0])
        return self._headers[key]

    def set_header(self, key: str, value: str, /) -> None:
        self._headers[key] = value

    def set_headers(self, values: dict[str, str], /) -> None:
        self._headers = self._headers | values

    def has_header(self, key: str, /) -> bool:
        return key in self._headers

    def set_content_type(self, value: str, /) -> None:
        self.set_header(HttpHeaders.ContentType, value)



from collections.abc import Callable
from typing import Protocol

from bolinette.core import Cache, __user_cache__, meta
from bolinette.web.ws import WebSocketSubResult, WebSocketSubscription


class WebSocketTopic[**SubP](Protocol):
    async def subscribe(
        self,
        sub: WebSocketSubscription,
        /,
        *args: SubP.args,
        **kwargs: SubP.kwargs,
    ) -> WebSocketSubResult: ...


class WebSocketTopicMeta:
    def __init__(self, name: str) -> None:
        self.name = name


def topic[**SubP](
    name: str, *, cache: Cache | None = None
) -> Callable[[type[WebSocketTopic[SubP]]], type[WebSocketTopic[SubP]]]:
    def decorator(cls: type[WebSocketTopic[SubP]]) -> type[WebSocketTopic[SubP]]:
        (cache or __user_cache__).add(WebSocketTopic, cls)
        meta.set(cls, WebSocketTopicMeta(name))
        return cls

    return decorator



from bolinette.web.ws.sub import (
    WebSocketSubscription as WebSocketSubscription,
    WebSocketSubResult as WebSocketSubResult,
)
from bolinette.web.ws.topic import WebSocketTopic as WebSocketTopic, topic as topic
from bolinette.web.ws.channel import channel as channel, ChannelMessage as ChannelMessage
from bolinette.web.ws.handler import WebSocketHandler as WebSocketHandler, WebSocketContext as WebSocketContext



import json
from collections.abc import Awaitable, Callable
from typing import Any, overload

from bolinette.web.abstract import ResponseState
from bolinette.web.asgi.types import HttpResponseResult, WebSocketSendResult
from bolinette.web.exceptions import InternalServerError


class AsgiResponse:
    def __init__(self, send: Callable[[HttpResponseResult], Awaitable[None]]) -> None:
        self._send = send
        self._status: int = 200
        self._headers: dict[str, str] = {}
        self._state: ResponseState = ResponseState.Idle

    @property
    def status(self) -> int:
        return self._status

    @property
    def headers(self) -> dict[str, str]:
        return {**self._headers}

    @property
    def state(self) -> ResponseState:
        return self._state

    async def open(self) -> None:
        if self._state != ResponseState.Idle:
            raise InternalServerError("Response already started")
        await self._send(
            {
                "type": "http.response.start",
                "status": self._status,
                "headers": [(k.encode(), v.encode()) for k, v in self._headers.items()],
            }
        )
        self._state = ResponseState.Started

    async def close(self) -> None:
        match self._state:
            case ResponseState.Idle:
                raise InternalServerError("Response has not started")
            case ResponseState.Closed:
                raise InternalServerError("Response has been closed")
            case ResponseState.Sending | ResponseState.Started:
                await self._send({"type": "http.response.body", "body": b"", "more_body": False})
                self._state = ResponseState.Closed

    async def write(self, raw: bytes) -> None:
        match self._state:
            case ResponseState.Sending:
                await self._send({"type": "http.response.body", "body": raw, "more_body": True})
            case ResponseState.Started:
                await self._send({"type": "http.response.body", "body": raw, "more_body": True})
                self._state = ResponseState.Sending
            case ResponseState.Idle:
                raise InternalServerError("Response has not started")
            case ResponseState.Closed:
                raise InternalServerError("Response has been closed")

    def set_status(self, status: int, /) -> None:
        self._status = status

    def set_header(self, key: str, value: str, /) -> None:
        self._headers[key] = value

    def has_header(self, key: str, /) -> bool:
        return key in self._headers

    def unset_header(self, key: str, /) -> None:
        if key in self._headers:
            del self._headers[key]


class AsgiSocketResponse:
    def __init__(self, send: Callable[[WebSocketSendResult], Awaitable[None]]) -> None:
        self._send = send

    @overload
    async def send(self, *, raw: bytes) -> None: ...
    @overload
    async def send(self, *, text: str) -> None: ...
    @overload
    async def send(self, *, json: Any, encoder: Callable[[object], str] = json.dumps) -> None: ...

    async def send(self, **kwargs: Any):
        if "raw" in kwargs:
            self._send({"type": "websocket.send", "bytes": kwargs["raw"]})
        if "text" in kwargs:
            self._send({"type": "websocket.send", "text": kwargs["text"]})
        if "json" in kwargs:
            encoder: Callable[[object], str] = kwargs.get("encoder", json.dumps)
            await self._send({"type": "websocket.send", "text": encoder(kwargs["json"])})



from http import HTTPStatus
from typing import Any

from bolinette.core.injection.resolver import ArgResolverOptions
from bolinette.core.mapping import Mapper
from bolinette.core.mapping.exceptions import (
    ConvertionError,
    DestinationNotNullableError,
    MappingError,
    SourceNotFoundError,
    ValidationError,
)
from bolinette.web import Payload
from bolinette.web.abstract import Request
from bolinette.web.exceptions import (
    BadRequestError,
    GroupedWebError,
    MissingParameterError,
    ParameterNotNullableError,
    WebError,
    WrongParameterTypeError,
)
from bolinette.web.routing import Route


class RouteParamArgResolver:
    def __init__(self, route: Route[..., Any], request: Request) -> None:
        self.route = route
        self.request = request

    def supports(self, options: ArgResolverOptions) -> bool:
        return options.name in self.request.path_params

    def resolve(self, options: ArgResolverOptions) -> tuple[str, Any]:
        value = self.request.path_params[options.name]
        t = options.t
        match t.cls:
            case cls if cls is int:
                value = int(value)
            case cls if cls is float:
                value = float(value)
            case cls if cls is bool:
                value = bool(value)
            case cls if cls is not str:
                raise Exception()  # TODO: Lever une vraie exception
            case _:
                pass
        return (options.name, value)


class RoutePayloadArgResolver:
    def __init__(self, route: Route[..., Any], mapper: Mapper, body: object | None) -> None:
        self.route = route
        self.mapper = mapper
        self.body = body

    def supports(self, options: ArgResolverOptions) -> bool:
        return any(
            isinstance(a, Payload) or (isinstance(a, type) and issubclass(a, Payload)) for a in options.t.annotated
        )

    def resolve(self, options: ArgResolverOptions) -> tuple[str, Any]:
        if self.body is None:
            if options.t.nullable:
                return options.name, None
            raise BadRequestError(
                "Payload expected but none provided",
                "payload.expected",
                ctrl=self.route.controller,
                route=self.route.func,
            )
        try:
            payload = self.mapper.map(type(self.body), options.t.cls, self.body, validate=True)
        except ValidationError as err:
            raise GroupedWebError(
                [self._transform_error(e) for e in err.errors],
                HTTPStatus.BAD_REQUEST,
                ctrl=self.route.controller,
                route=self.route.func,
            ) from err
        return options.name, payload

    def _transform_error(self, err: MappingError) -> WebError:
        match err:
            case SourceNotFoundError():
                return MissingParameterError(
                    err.dest,
                    ctrl=self.route.controller,
                    route=self.route.func,
                )
            case DestinationNotNullableError():
                return ParameterNotNullableError(
                    err.dest,
                    ctrl=self.route.controller,
                    route=self.route.func,
                )
            case ConvertionError():
                return WrongParameterTypeError(
                    err.dest,
                    err.target,
                    ctrl=self.route.controller,
                    route=self.route.func,
                )
            case _:
                raise NotImplementedError(type(err), err) from err



from bolinette.web.abstract.request import Request as Request, WebSocketRequest as WebSocketRequest
from bolinette.web.abstract.response import (
    Response as Response,
    WebSocketResponse as WebSocketResponse,
    ResponseState as ResponseState,
)



from enum import StrEnum


class HttpHeaders(StrEnum):
    AIM = "A-IM"
    Accept = "Accept"
    AcceptAdditions = "Accept-Additions"
    AcceptCH = "Accept-CH"
    AcceptCharset = "Accept-Charset"
    AcceptDatetime = "Accept-Datetime"
    AcceptEncoding = "Accept-Encoding"
    AcceptFeatures = "Accept-Features"
    AcceptLanguage = "Accept-Language"
    AcceptPatch = "Accept-Patch"
    AcceptPost = "Accept-Post"
    AcceptRanges = "Accept-Ranges"
    AcceptSignature = "Accept-Signature"
    AccessControl = "Access-Control"
    AccessControlAllowCredentials = "Access-Control-Allow-Credentials"
    AccessControlAllowHeaders = "Access-Control-Allow-Headers"
    AccessControlAllowMethods = "Access-Control-Allow-Methods"
    AccessControlAllowOrigin = "Access-Control-Allow-Origin"
    AccessControlExposeHeaders = "Access-Control-Expose-Headers"
    AccessControlMaxAge = "Access-Control-Max-Age"
    AccessControlRequestHeaders = "Access-Control-Request-Headers"
    AccessControlRequestMethod = "Access-Control-Request-Method"
    Age = "Age"
    Allow = "Allow"
    ALPN = "ALPN"
    AltSvc = "Alt-Svc"
    AltUsed = "Alt-Used"
    Alternates = "Alternates"
    AMPCacheTransform = "AMP-Cache-Transform"
    ApplyToRedirectRef = "Apply-To-Redirect-Ref"
    AuthenticationControl = "Authentication-Control"
    AuthenticationInfo = "Authentication-Info"
    Authorization = "Authorization"
    CExt = "C-Ext"
    CMan = "C-Man"
    COpt = "C-Opt"
    CPEP = "C-PEP"
    CPEPInfo = "C-PEP-Info"
    CacheControl = "Cache-Control"
    CacheStatus = "Cache-Status"
    CalManagedID = "Cal-Managed-ID"
    CalDAVTimezones = "CalDAV-Timezones"
    CapsuleProtocol = "Capsule-Protocol"
    CDNCacheControl = "CDN-Cache-Control"
    CDNLoop = "CDN-Loop"
    CertNotAfter = "Cert-Not-After"
    CertNotBefore = "Cert-Not-Before"
    ClearSiteData = "Clear-Site-Data"
    ClientCert = "Client-Cert"
    ClientCertChain = "Client-Cert-Chain"
    Close = "Close"
    ConfigurationContext = "Configuration-Context"
    Connection = "Connection"
    ContentBase = "Content-Base"
    ContentDigest = "Content-Digest"
    ContentDisposition = "Content-Disposition"
    ContentEncoding = "Content-Encoding"
    ContentID = "Content-ID"
    ContentLanguage = "Content-Language"
    ContentLength = "Content-Length"
    ContentLocation = "Content-Location"
    ContentMD5 = "Content-MD5"
    ContentRange = "Content-Range"
    ContentScriptType = "Content-Script-Type"
    ContentSecurityPolicy = "Content-Security-Policy"
    ContentSecurityPolicyReportOnly = "Content-Security-Policy-Report-Only"
    ContentStyleType = "Content-Style-Type"
    ContentType = "Content-Type"
    ContentVersion = "Content-Version"
    Cookie = "Cookie"
    Cookie2 = "Cookie2"
    CrossOriginEmbedderPolicy = "Cross-Origin-Embedder-Policy"
    CrossOriginEmbedderPolicyReportOnly = "Cross-Origin-Embedder-Policy-Report-Only"
    CrossOriginOpenerPolicy = "Cross-Origin-Opener-Policy"
    CrossOriginOpenerPolicyReportOnly = "Cross-Origin-Opener-Policy-Report-Only"
    CrossOriginResourcePolicy = "Cross-Origin-Resource-Policy"
    DASL = "DASL"
    Date = "Date"
    DAV = "DAV"
    DefaultStyle = "Default-Style"
    DeltaBase = "Delta-Base"
    Depth = "Depth"
    DerivedFrom = "Derived-From"
    Destination = "Destination"
    DifferentialID = "Differential-ID"
    Digest = "Digest"
    DPoP = "DPoP"
    DPoPNonce = "DPoP-Nonce"
    EarlyData = "Early-Data"
    EDIINTFeatures = "EDIINT-Features"
    ETag = "ETag"
    Expect = "Expect"
    ExpectCT = "Expect-CT"
    Expires = "Expires"
    Ext = "Ext"
    Forwarded = "Forwarded"
    From = "From"
    GetProfile = "GetProfile"
    Hobareg = "Hobareg"
    Host = "Host"
    HTTP2Settings = "HTTP2-Settings"
    If = "If"
    IfMatch = "If-Match"
    IfModifiedSince = "If-Modified-Since"
    IfNoneMatch = "If-None-Match"
    IfRange = "If-Range"
    IfScheduleTagMatch = "If-Schedule-Tag-Match"
    IfUnmodifiedSince = "If-Unmodified-Since"
    IM = "IM"
    IncludeReferredTokenBindingID = "Include-Referred-Token-Binding-ID"
    Isolation = "Isolation"
    KeepAlive = "Keep-Alive"
    Label = "Label"
    LastEventID = "Last-Event-ID"
    LastModified = "Last-Modified"
    Link = "Link"
    Location = "Location"
    LockToken = "Lock-Token"
    Man = "Man"
    MaxForwards = "Max-Forwards"
    MementoDatetime = "Memento-Datetime"
    Meter = "Meter"
    MethodCheck = "Method-Check"
    MethodCheckExpires = "Method-Check-Expires"
    MIMEVersion = "MIME-Version"
    Negotiate = "Negotiate"
    NEL = "NEL"
    ODataEntityId = "OData-EntityId"
    ODataIsolation = "OData-Isolation"
    ODataMaxVersion = "OData-MaxVersion"
    ODataVersion = "OData-Version"
    Opt = "Opt"
    OptionalWWWAuthenticate = "Optional-WWW-Authenticate"
    OrderingType = "Ordering-Type"
    Origin = "Origin"
    OriginAgentCluster = "Origin-Agent-Cluster"
    OSCORE = "OSCORE"
    OSLCCoreVersion = "OSLC-Core-Version"
    Overwrite = "Overwrite"
    P3P = "P3P"
    PEP = "PEP"
    PEPInfo = "PEP-Info"
    PermissionsPolicy = "Permissions-Policy"
    PICSLabel = "PICS-Label"
    PingFrom = "Ping-From"
    PingTo = "Ping-To"
    Position = "Position"
    Pragma = "Pragma"
    Prefer = "Prefer"
    PreferenceApplied = "Preference-Applied"
    Priority = "Priority"
    ProfileObject = "ProfileObject"
    Protocol = "Protocol"
    ProtocolInfo = "Protocol-Info"
    ProtocolQuery = "Protocol-Query"
    ProtocolRequest = "Protocol-Request"
    ProxyAuthenticate = "Proxy-Authenticate"
    ProxyAuthenticationInfo = "Proxy-Authentication-Info"
    ProxyAuthorization = "Proxy-Authorization"
    ProxyFeatures = "Proxy-Features"
    ProxyInstruction = "Proxy-Instruction"
    ProxyStatus = "Proxy-Status"
    Public = "Public"
    PublicKeyPins = "Public-Key-Pins"
    PublicKeyPinsReportOnly = "Public-Key-Pins-Report-Only"
    Range = "Range"
    RedirectRef = "Redirect-Ref"
    Referer = "Referer"
    RefererRoot = "Referer-Root"
    Refresh = "Refresh"
    RepeatabilityClientID = "Repeatability-Client-ID"
    RepeatabilityFirstSent = "Repeatability-First-Sent"
    RepeatabilityRequestID = "Repeatability-Request-ID"
    RepeatabilityResult = "Repeatability-Result"
    ReplayNonce = "Replay-Nonce"
    ReportingEndpoints = "Reporting-Endpoints"
    ReprDigest = "Repr-Digest"
    RetryAfter = "Retry-After"
    Safe = "Safe"
    ScheduleReply = "Schedule-Reply"
    ScheduleTag = "Schedule-Tag"
    SecGPC = "Sec-GPC"
    SecPurpose = "Sec-Purpose"
    SecTokenBinding = "Sec-Token-Binding"
    SecWebSocketAccept = "Sec-WebSocket-Accept"
    SecWebSocketExtensions = "Sec-WebSocket-Extensions"
    SecWebSocketKey = "Sec-WebSocket-Key"
    SecWebSocketProtocol = "Sec-WebSocket-Protocol"
    SecWebSocketVersion = "Sec-WebSocket-Version"
    SecurityScheme = "Security-Scheme"
    Server = "Server"
    ServerTiming = "Server-Timing"
    SetCookie = "Set-Cookie"
    SetCookie2 = "Set-Cookie2"
    SetProfile = "SetProfile"
    Signature = "Signature"
    SignatureInput = "Signature-Input"
    SLUG = "SLUG"
    SoapAction = "SoapAction"
    StatusURI = "Status-URI"
    StrictTransportSecurity = "Strict-Transport-Security"
    Sunset = "Sunset"
    SurrogateCapability = "Surrogate-Capability"
    SurrogateControl = "Surrogate-Control"
    TCN = "TCN"
    TE = "TE"
    Timeout = "Timeout"
    TimingAllowOrigin = "Timing-Allow-Origin"
    Topic = "Topic"
    Traceparent = "Traceparent"
    Tracestate = "Tracestate"
    Trailer = "Trailer"
    TransferEncoding = "Transfer-Encoding"
    TTL = "TTL"
    Upgrade = "Upgrade"
    Urgency = "Urgency"
    URI = "URI"
    UserAgent = "User-Agent"
    VariantVary = "Variant-Vary"
    Vary = "Vary"
    Via = "Via"
    WantContentDigest = "Want-Content-Digest"
    WantDigest = "Want-Digest"
    WantReprDigest = "Want-Repr-Digest"
    Warning = "Warning"
    WWWAuthenticate = "WWW-Authenticate"
    XContentTypeOptions = "X-Content-Type-Options"
    XFrameOptions = "X-Frame-Options"



from collections.abc import Awaitable, Callable
from typing import Any

from bolinette.core.bolinette import Bolinette
from bolinette.web.asgi import (
    AsgiCallable,
    AsgiRequest,
    AsgiResponse,
    AsgiSocketRequest,
    AsgiSocketResponse,
    HttpReceivedEvent,
    HttpRequestEvent,
    HttpResponseResult,
    HttpScope,
    LifespanReceivedEvent,
    LifespanResult,
    LifespanShutdownResult,
    LifespanStartupResult,
    Scope,
    WebSocketConnectResult,
    WebSocketReceivedEvent,
    WebSocketResult,
    WebSocketScope,
)
from bolinette.web.resources import WebResources
from bolinette.web.ws import WebSocketHandler


class AsgiApplication:
    def __init__(self, blnt: Bolinette) -> None:
        self._blnt = blnt
        self._resources: WebResources | None = None
        self._ws_handler: WebSocketHandler | None = None

    async def _handle_startup(
        self,
        send: Callable[[LifespanStartupResult], Awaitable[None]],
    ) -> None:
        try:
            await self._blnt.startup()
            self._blnt.injection.add(AsgiApplication, "singleton", instance=self)
            await send({"type": "lifespan.startup.complete"})
        except BaseException:
            await send({"type": "lifespan.startup.failed"})

    async def _handle_shutdown(
        self,
        send: Callable[[LifespanShutdownResult], Awaitable[None]],
    ) -> None:
        try:
            await send({"type": "lifespan.shutdown.complete"})
        except BaseException:
            await send({"type": "lifespan.shutdown.failed"})

    async def _handle_lifespan(
        self,
        receive: Callable[[], Awaitable[LifespanReceivedEvent]],
        send: Callable[[LifespanResult], Awaitable[None]],
    ) -> None:
        received = await receive()
        match received["type"]:
            case "lifespan.startup":
                await self._handle_startup(send)
            case "lifespan.shutdown":
                await self._handle_shutdown(send)

    async def _handle_http_request(
        self,
        scope: HttpScope,
        received: HttpRequestEvent,
        receive: Callable[[], Awaitable[HttpReceivedEvent]],
        send: Callable[[HttpResponseResult], Awaitable[None]],
    ) -> None:
        if self._resources is None:
            self._resources = self._blnt.injection.require(WebResources)

        request = AsgiRequest(scope["method"], scope["path"], {}, {}, received, receive)
        response = AsgiResponse(send)
        await self._resources.dispatch(request, response)

    async def _handle_http(
        self,
        scope: HttpScope,
        receive: Callable[[], Awaitable[HttpReceivedEvent]],
        send: Callable[[HttpResponseResult], Awaitable[None]],
    ) -> None:
        received = await receive()
        match received["type"]:
            case "http.request":
                await self._handle_http_request(scope, received, receive, send)
            case "http.disconnect":
                raise NotImplementedError()

    async def _handle_ws_connect(
        self,
        send: Callable[[WebSocketConnectResult], Awaitable[None]],
    ) -> None:
        try:
            await send({"type": "websocket.accept"})
        except BaseException:
            await send({"type": "websocket.close"})

    async def _handle_ws(
        self,
        scope: WebSocketScope,
        receive: Callable[[], Awaitable[WebSocketReceivedEvent]],
        send: Callable[[WebSocketResult], Awaitable[None]],
    ) -> None:
        response = AsgiSocketResponse(send)
        if self._ws_handler is None:
            self._ws_handler = self._blnt.injection.require(WebSocketHandler)
        while True:
            received = await receive()
            match received["type"]:
                case "websocket.connect":
                    await self._handle_ws_connect(send)
                case "websocket.receive":
                    request = AsgiSocketRequest(received.get("bytes", None), received.get("text", None))
                    await self._ws_handler.handle(request, response)
                case "websocket.disconnect":
                    await self._ws_handler.remove_connection(response)
                    break

    def get_app(self) -> AsgiCallable:
        async def app(
            scope: Scope,
            receive: Callable[[], Awaitable[Any]],
            send: Callable[[Any], Awaitable[None]],
        ) -> None:
            match scope["type"]:
                case "lifespan":
                    await self._handle_lifespan(receive, send)
                case "http":
                    await self._handle_http(scope, receive, send)
                case "websocket":
                    await self._handle_ws(scope, receive, send)

        return app



from typing import Any, Literal, TypedDict

type SocketContent = int | float | bool | str | list[SocketContent] | dict[str, SocketContent]


class ChannelSubscribeRequest(TypedDict):
    action: Literal["sub"]
    topic: str
    channel: str


class ChannelUnsubscribeRequest(TypedDict):
    action: Literal["unsub"]
    topic: str
    channel: str


class ChannelSendRequest[D: SocketContent](TypedDict):
    action: Literal["send"]
    topic: str
    channel: str
    data: D


class ChannelCloseRequest(TypedDict):
    action: Literal["close"]


type ChannelRequest = (
    ChannelSubscribeRequest | ChannelUnsubscribeRequest | ChannelSendRequest[Any] | ChannelCloseRequest
)



from typing import Any

from bolinette.web.abstract import Request
from bolinette.web.exceptions import MethodNotAllowedDispatchError, NotFoundDispatchError
from bolinette.web.routing import Resource, ResourceNode, Route


class Router:
    def __init__(self) -> None:
        self.root_node: ResourceNode | None = None

    def add_route(self, route: Route[..., Any], /) -> None:
        new_node = ResourceNode.from_resource(Resource(route.path, [route]))
        if self.root_node is None:
            self.root_node = new_node
        else:
            self.root_node = ResourceNode.merge(self.root_node, new_node)

    def dispatch(self, request: Request) -> Route[..., Any]:
        def find_node(req_path: str, res_node: ResourceNode | None) -> ResourceNode | None:
            if not res_node or not len(req_path):
                return None
            length = res_node.match(req_path, request.path_params)
            if length <= 0:
                return None
            if length == len(req_path):
                return res_node
            req_path = req_path[length:]
            for subnode in res_node.subnodes:
                res = find_node(req_path, subnode)
                if res is not None:
                    return res
            return None

        node = find_node(request.path, self.root_node)
        if not node or not node.resource:
            raise NotFoundDispatchError(request.path)
        if request.method not in node.resource:
            raise MethodNotAllowedDispatchError(request.path)
        return node.resource[request.method]



from typing import Self, override

from bolinette import core
from bolinette.core import Cache, startup
from bolinette.core.extension import Extension
from bolinette.core.injection import Injection, injectable
from bolinette.web.config import WebConfig
from bolinette.web.resources import WebResources
from bolinette.web.ws import WebSocketHandler


class WebExtension(Extension):
    def __init__(self) -> None:
        super().__init__("web", [core])
        self.config = WebConfig()

    @override
    def add_cached(self, cache: Cache) -> None:
        injectable(strategy="singleton", cache=cache)(WebResources)
        injectable(strategy="singleton", cache=cache)(WebSocketHandler)

        startup(cache=cache)(init_web_ext)

    def use_sockets(self) -> Self:
        self.config.use_sockets = True
        return self


web_ext = WebExtension()


async def init_web_ext(inject: Injection) -> None:
    inject.add(WebExtension, strategy="singleton", instance=web_ext)
    inject.add(WebConfig, strategy="singleton", instance=web_ext.config)



import re
from collections.abc import Callable
from typing import Concatenate

from bolinette.core import meta
from bolinette.web.abstract import WebSocketResponse
from bolinette.web.ws import WebSocketTopic
from bolinette.web.ws.requests import SocketContent


class ChannelMessage[T: SocketContent]:
    def __init__(self, channel: str, value: T, response: WebSocketResponse) -> None:
        self.channel = channel
        self.type = type(value)
        self.value = value
        self.response = response


type ChannelFunc[W: WebSocketTopic[...], M: SocketContent, **P, T] = Callable[Concatenate[W, ChannelMessage[M], P], T]


class WebSocketChannelMeta[W: WebSocketTopic[...], M: SocketContent, **P, T]:
    def __init__(self, pattern: str, func: ChannelFunc[W, M, P, T]) -> None:
        self.pattern = re.compile(pattern)
        self.func = func


def channel[W: WebSocketTopic[...], M: SocketContent, **P, T](
    pattern: str,
    /,
) -> Callable[[ChannelFunc[W, M, P, T]], ChannelFunc[W, M, P, T]]:
    def decorator(func: ChannelFunc[W, M, P, T]) -> ChannelFunc[W, M, P, T]:
        meta.set(func, WebSocketChannelMeta(pattern, func))
        return func

    return decorator
